# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from flask import request
from flask_restful import Resource
from taipy.config.config import Config
from taipy.core.exceptions.exceptions import NonExistingScenario, NonExistingScenarioConfig
from taipy.core.scenario._scenario_manager_factory import _ScenarioManagerFactory

from ...commons.to_from_model import _to_model
from ..exceptions.exceptions import ConfigIdMissingException
from ..middlewares._middleware import _middleware
from ..schemas import ScenarioResponseSchema


def _get_or_raise(scenario_id: str):
    manager = _ScenarioManagerFactory._build_manager()
    scenario = manager._get(scenario_id)
    if scenario is None:
        raise NonExistingScenario(scenario_id)
    return scenario


REPOSITORY = "scenario"


class ScenarioResource(Resource):
    """Single object resource

    ---
    get:
      tags:
        - api
      description: |
        Returns a `ScenarioSchema^` representing the unique scenario identified by *scenario_id*. If no scenario
        corresponds to *scenario_id*, a `404` error is returned.

        !!! Example

            === "Curl"
                ```shell
                    curl -X GET http://localhost:5000/api/v1/scenarios/SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c
                ```
                In this example the REST API is served on port 5000 on localhost. We are using curl command line
                client.

                `SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c` is the value of the *scenario_id* parameter. It
                represents the identifier of the Scenario we want to retrieve.

                In case of success here is an example of the response:
                ``` JSON
                {"scenario": {
                    "cycle": "CYCLE_863418_fdd1499a-8925-4540-93fd-9dbfb4f0846d",
                    "id": "SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c",
                    "properties": {},
                    "tags": [],
                    "version": "latest",
                    "pipelines": [
                        "PIPELINE_mean_baseline_5af317c9-34df-48b4-8a8a-bf4007e1de99",
                        "PIPELINE_arima_90aef6b9-8922-4a0c-b625-b2c6f3d19fa4"],
                    "subscribers": [],
                    "creation_date": "2022-08-15T19:21:01.871587",
                    "primary_scenario": true}}
                ```

                In case of failure here is an example of the response:
                ``` JSON
                {"message": "SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c not found."}
                ```

            === "Python"
                This Python example requires the 'requests' package to be installed (`pip install requests`).
                ```python
                import requests
                    response = requests.get(
                    "http://localhost:5000/api/v1/scenarios/SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c")
                    print(response)
                    print(response.json())
                ```
                `SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c` is the value of the  *scenario_id* parameter. It
                represents the identifier of the Cycle we want to retrieve.

                In case of success here is an output example:
                ```
                <Response [200]>
                {"scenario": {
                    "cycle": "CYCLE_863418_fdd1499a-8925-4540-93fd-9dbfb4f0846d",
                    "id": "SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c",
                    "properties": {},
                    "tags": [],
                    "version": "latest",
                    "pipelines": [
                        "PIPELINE_mean_baseline_5af317c9-34df-48b4-8a8a-bf4007e1de99",
                        "PIPELINE_arima_90aef6b9-8922-4a0c-b625-b2c6f3d19fa4"],
                    "subscribers": [],
                    "creation_date": "2022-08-15T19:21:01.871587",
                    "primary_scenario": true}}
                ```

                In case of failure here is an output example:
                ```
                <Response [404]>
                {'message': 'Scenario SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c not found.'}

                ```

        !!! Note
            When the authorization feature is activated (available in Taipy Enterprise edition only), this endpoint
            requires the `TAIPY_READER` role.

        parameters:
        - in: path
          name: scenario_id
          schema:
            type: string
          description: The identifier of the scenario to retrieve.
      responses:
        200:
          content:
            application/json:
              schema:
                type: object
                properties:
                  scenario: ScenarioSchema
        404:
          description: No scenario has the *scenario_id* identifier.
    delete:
      tags:
        - api
      description: |
        Delete the `Scenario^` scenario identified by the *scenario_id* given as parameter. If the scenario does not
        exist, a 404 error is returned.

        !!! Example

            === "Curl"
                ```shell
                    curl -X DELETE http://localhost:5000/api/v1/scenarios/SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c
                ```
                In this example the REST API is served on port 5000 on localhost. We are using curl command line
                client.

                `SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c` is the value of the  *scenario_id* parameter. It
                represents the identifier of the scenario we want to delete.

                In case of success here is an example of the response:
                ``` JSON
                {"msg": "Scenario SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c deleted."}
                ```

                In case of failure here is an example of the response:
                ``` JSON
                {"message": "Scenario SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c not found."}
                ```

            === "Python"
                This Python example requires the 'requests' package to be installed (`pip install requests`).
                ```python
                import requests
                    response = requests.delete(
                    "http://localhost:5000/api/v1/scenarios/SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c")
                    print(response)
                    print(response.json())
                ```
                `SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c` is the value of the *scenario_id* parameter. It
                represents the identifier of the Scenario we want to delete.

                In case of success here is an output example:
                ```
                <Response [200]>
                {"msg": "Scenario SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c deleted."}
                ```

                In case of failure here is an output example:
                ```
                <Response [404]>
                {'message': 'Scenario SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c not found.'}

                ```

        !!! Note
            When the authorization feature is activated (available in Taipy Enterprise edition only), this endpoint
            requires the `TAIPY_EDITOR` role.

      parameters:
        - in: path
          name: scenario_id
          schema:
            type: string
          description: The identifier of the scenario to delete.
      responses:
        200:
          content:
            application/json:
              schema:
                type: object
                properties:
                  message:
                    type: string
                    description: Status message.
        404:
          description: No scenario has the *scenario_id* identifier.
    """

    def __init__(self, **kwargs):
        self.logger = kwargs.get("logger")

    @_middleware
    def get(self, scenario_id):
        schema = ScenarioResponseSchema()
        scenario = _get_or_raise(scenario_id)
        return {"scenario": schema.dump(_to_model(REPOSITORY, scenario))}

    @_middleware
    def delete(self, scenario_id):
        manager = _ScenarioManagerFactory._build_manager()
        _get_or_raise(scenario_id)
        manager._delete(scenario_id)
        return {"message": f"Scenario {scenario_id} was deleted."}


class ScenarioList(Resource):
    """Creation and get_all

    ---
    get:
      tags:
        - api
      summary: Get all scenarios.
      description: |
        Returns a `ScenarioSchema^` list representing all existing Scenarios.

        !!! Example

            === "Curl"
                ```shell
                    curl -X GET http://localhost:5000/api/v1/scenarios
                ```
                In this example the REST API is served on port 5000 on localhost. We are using curl command line
                client.

                Here is an example of the response:
                ``` JSON
                [{
                    "cycle": "CYCLE_863418_fdd1499a-8925-4540-93fd-9dbfb4f0846d",
                    "id": "SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c",
                    "properties": {},
                    "tags": [],
                    "version": "latest",
                    "pipelines": [
                        "PIPELINE_mean_baseline_5af317c9-34df-48b4-8a8a-bf4007e1de99",
                        "PIPELINE_arima_90aef6b9-8922-4a0c-b625-b2c6f3d19fa4"],
                    "subscribers": [],
                    "creation_date": "2022-08-15T19:21:01.871587",
                    "primary_scenario": true
                    }
                ]
                ```

                If there is no scenario, the response is an empty list as follows:
                ``` JSON
                []
                ```

            === "Python"
                This Python example requires the 'requests' package to be installed (`pip install requests`).
                ```python
                import requests
                    response = requests.get("http://localhost:5000/api/v1/scenarios")
                    print(response)
                    print(response.json())
                ```

                In case of success here is an output example:
                ```
                <Response [200]>
                [{
                    "cycle": "CYCLE_863418_fdd1499a-8925-4540-93fd-9dbfb4f0846d",
                    "id": "SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c",
                    "properties": {},
                    "tags": [],
                    "version": "latest",
                    "pipelines": [
                        "PIPELINE_mean_baseline_5af317c9-34df-48b4-8a8a-bf4007e1de99",
                        "PIPELINE_arima_90aef6b9-8922-4a0c-b625-b2c6f3d19fa4"],
                    "subscribers": [],
                    "creation_date": "2022-08-15T19:21:01.871587",
                    "primary_scenario": true
                    }
                ]
                ```

                If there is no scenario, the response is an empty list as follows:
                ```
                <Response [200]>
                []
                ```

        !!! Note
            When the authorization feature is activated (available in Taipy Enterprise edition only), this endpoint
            requires the `TAIPY_READER` role.

      responses:
        200:
          content:
            application/json:
              schema:
                allOf:
                  - type: object
                    properties:
                      results:
                        type: array
                        items:
                          $ref: '#/components/schemas/ScenarioSchema'
    post:
      tags:
        - api
      description: |
        Creates a new scenario from the  *config_id*. If the config does not exist, a 404 error is returned.

        !!! Example

            === "Curl"
                ```shell
                    curl -X POST http://localhost:5000/api/v1/scenarios?config_id=my_scenario_config
                ```
                In this example the REST API is served on port 5000 on localhost. We are using curl command line
                client.

                In this example the *config_id* value ("my_scenario_config") is given as parameter directly in the
                url. A corresponding `ScenarioConfig^` must exist and must have been configured before.

                Here is the output message example:
                ```
                {"msg": "scenario created.",
                "scenario": {
                    "cycle": "CYCLE_863418_fdd1499a-8925-4540-93fd-9dbfb4f0846d",
                    "id": "SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c",
                    "properties": {},
                    "tags": [],
                    "version": "latest",
                    "pipelines": [
                        "PIPELINE_mean_baseline_5af317c9-34df-48b4-8a8a-bf4007e1de99",
                        "PIPELINE_arima_90aef6b9-8922-4a0c-b625-b2c6f3d19fa4"],
                    "subscribers": [],
                    "creation_date": "2022-08-15T19:21:01.871587",
                    "primary_scenario": true}
                }
                ```

            === "Python"
                This Python example requires the 'requests' package to be installed (`pip install requests`).
                ```python
                import requests
                    response = requests.post("http://localhost:5000/api/v1/scenarios?config_id=my_scenario_config")
                    print(response)
                    print(response.json())
                ```
                In this example the *config_id* value ("my_scenario_config") is given as parameter directly in the
                url. A corresponding `ScenarioConfig^` must exist and must have been configured before.

                Here is the output example:
                ```
                <Response [201]>
                {"msg": "scenario created.",
                "scenario": {
                    "cycle": "CYCLE_863418_fdd1499a-8925-4540-93fd-9dbfb4f0846d",
                    "id": "SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c",
                    "properties": {},
                    "tags": [],
                    "version": "latest",
                    "pipelines": [
                        "PIPELINE_mean_baseline_5af317c9-34df-48b4-8a8a-bf4007e1de99",
                        "PIPELINE_arima_90aef6b9-8922-4a0c-b625-b2c6f3d19fa4"],
                    "subscribers": [],
                    "creation_date": "2022-08-15T19:21:01.871587",
                    "primary_scenario": true}
                }
                ```

        !!! Note
            When the authorization feature is activated (available in Taipy Enterprise edition only), this endpoint
            requires the `TAIPY_EDITOR` role.

      parameters:
        - in: query
          name: config_id
          schema:
            type: string
          description: The identifier of the scenario configuration.
      responses:
        201:
          content:
            application/json:
              schema:
                type: object
                properties:
                  message:
                    type: string
                    description: Status message.
                  scenario: ScenarioSchema
    """

    def __init__(self, **kwargs):
        self.logger = kwargs.get("logger")

    def fetch_config(self, config_id):
        config = Config.scenarios.get(config_id)
        if not config:
            raise NonExistingScenarioConfig(config_id)
        return config

    @_middleware
    def get(self):
        schema = ScenarioResponseSchema(many=True)
        manager = _ScenarioManagerFactory._build_manager()
        scenarios = [_to_model(REPOSITORY, scenario) for scenario in manager._get_all()]
        return schema.dump(scenarios)

    @_middleware
    def post(self):
        args = request.args
        config_id = args.get("config_id")

        response_schema = ScenarioResponseSchema()
        manager = _ScenarioManagerFactory._build_manager()

        if not config_id:
            raise ConfigIdMissingException

        config = self.fetch_config(config_id)
        scenario = manager._create(config)

        return {
            "message": "Scenario was created.",
            "scenario": response_schema.dump(_to_model(REPOSITORY, scenario)),
        }, 201


class ScenarioExecutor(Resource):
    """Execute a scenario

    ---
    post:
      tags:
        - api
      description: |
        Executes a scenario by *scenario_id*. If the scenario does not exist, a 404 error is returned.

        !!! Example

            === "Curl"
                ```shell
                    curl -X POST http://localhost:5000/api/v1/scenarios/submit/SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c
                ```
                In this example the REST API is served on port 5000 on localhost. We are using curl command line
                client.

                `SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c` is the value of the *scenario_id* parameter. It
                represents the identifier of the Scenario we want to submit.

                Here is the output message example:
                ```
                {"message": "Executed scenario SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c."}
                ```

            === "Python"
                This Python example requires the 'requests' package to be installed (`pip install requests`).
                ```python
                    import requests
                        response = requests.post(
                        "http://localhost:5000/api/v1/scenarios/submit/SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c")
                        print(response)
                        print(response.json())
                ```
                `SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c` is the value of the *scenario_id* parameter. It
                represents the identifier of the Scenario we want to submit.

                Here is the output example:
                ```
                <Response [202]>
                {"message": "Executed scenario SCENARIO_63cb358d-5834-4d73-84e4-a6343df5e08c."}
                ```

        !!! Note
            When the authorization feature is activated (available in Taipy Enterprise edition only), this endpoint
            requires the `TAIPY_EXECUTOR` role.

      parameters:
        - in: path
          name: scenario_id
          schema:
            type: string
          description: The identifier of the scenario to submit.
      responses:
        202:
          content:
            application/json:
              schema:
                type: object
                properties:
                  message:
                    type: string
                    description: Status message.
                  scenario: ScenarioSchema
        404:
          description: No scenario has the *scenario_id* identifier.
    """

    def __init__(self, **kwargs):
        self.logger = kwargs.get("logger")

    @_middleware
    def post(self, scenario_id):
        _get_or_raise(scenario_id)
        manager = _ScenarioManagerFactory._build_manager()
        manager._submit(scenario_id)
        return {"message": f"Scenario {scenario_id} was submitted."}




# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from apispec import APISpec
from apispec.exceptions import APISpecError
from apispec.ext.marshmallow import MarshmallowPlugin
from apispec_webframeworks.flask import FlaskPlugin
from flask import Blueprint, jsonify, render_template


class FlaskRestfulPlugin(FlaskPlugin):
    """Small plugin override to handle flask-restful resources"""

    @staticmethod
    def _rule_for_view(view, app=None):
        view_funcs = app.view_functions
        endpoint = None

        for ept, view_func in view_funcs.items():
            if hasattr(view_func, "view_class"):
                view_func = view_func.view_class

            if view_func == view:
                endpoint = ept

        if not endpoint:
            raise APISpecError("Could not find endpoint for view {0}".format(view))

        # WARNING: Assume 1 rule per view function for now
        rule = app.url_map._rules_by_endpoint[endpoint][0]
        return rule


class APISpecExt:
    """Very simple and small extension to use apispec with this API as a flask extension"""

    def __init__(self, app=None, **kwargs):
        self.spec = None

        if app is not None:
            self.init_app(app, **kwargs)

    def init_app(self, app, **kwargs):
        app.config.setdefault("APISPEC_TITLE", "Taipy Rest")
        app.config.setdefault("APISPEC_VERSION", "1.0.0")
        app.config.setdefault("OPENAPI_VERSION", "3.0.2")
        app.config.setdefault("SWAGGER_JSON_URL", "/swagger.json")
        app.config.setdefault("SWAGGER_UI_URL", "/swagger-ui")
        app.config.setdefault("OPENAPI_YAML_URL", "/openapi.yaml")
        app.config.setdefault("REDOC_UI_URL", "/redoc-ui")
        app.config.setdefault("SWAGGER_URL_PREFIX", None)

        self.spec = APISpec(
            title=app.config["APISPEC_TITLE"],
            version=app.config["APISPEC_VERSION"],
            openapi_version=app.config["OPENAPI_VERSION"],
            plugins=[MarshmallowPlugin(), FlaskRestfulPlugin()],
            **kwargs
        )

        blueprint = Blueprint(
            "swagger",
            __name__,
            template_folder="./templates",
            url_prefix=app.config["SWAGGER_URL_PREFIX"],
        )

        blueprint.add_url_rule(app.config["SWAGGER_JSON_URL"], "swagger_json", self.swagger_json)
        blueprint.add_url_rule(app.config["SWAGGER_UI_URL"], "swagger_ui", self.swagger_ui)
        blueprint.add_url_rule(app.config["OPENAPI_YAML_URL"], "openapi_yaml", self.openapi_yaml)
        blueprint.add_url_rule(app.config["REDOC_UI_URL"], "redoc_ui", self.redoc_ui)

        app.register_blueprint(blueprint)

    def swagger_json(self):
        return jsonify(self.spec.to_dict())

    def swagger_ui(self):
        return render_template("swagger.j2")

    def openapi_yaml(self):
        # Manually inject ReDoc's Authentication legend, then remove it
        self.spec.tag(
            {
                "name": "authentication",
                "x-displayName": "Authentication",
                "description": "<SecurityDefinitions />",
            }
        )
        redoc_spec = self.spec.to_yaml()
        self.spec._tags.pop(0)
        return redoc_spec

    def redoc_ui(self):
        return render_template("redoc.j2")



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from flask import Blueprint, current_app
from flask_restful import Api

from taipy.core.common._utils import _load_fct
from taipy.logger._taipy_logger import _TaipyLogger

from ..extensions import apispec
from .middlewares._middleware import _using_enterprise
from .resources import (
    CycleList,
    CycleResource,
    DataNodeList,
    DataNodeReader,
    DataNodeResource,
    DataNodeWriter,
    JobExecutor,
    JobList,
    JobResource,
    PipelineExecutor,
    PipelineList,
    PipelineResource,
    ScenarioExecutor,
    ScenarioList,
    ScenarioResource,
    TaskExecutor,
    TaskList,
    TaskResource,
)
from .schemas import CycleSchema, DataNodeSchema, JobSchema, PipelineSchema, ScenarioSchema, TaskSchema

_logger = _TaipyLogger._get_logger()


blueprint = Blueprint("api", __name__, url_prefix="/api/v1")

api = Api(blueprint)

api.add_resource(
    DataNodeResource,
    "/datanodes/<string:datanode_id>",
    endpoint="datanode_by_id",
    resource_class_kwargs={"logger": _logger},
)

api.add_resource(
    DataNodeReader,
    "/datanodes/<string:datanode_id>/read",
    endpoint="datanode_reader",
    resource_class_kwargs={"logger": _logger},
)

api.add_resource(
    DataNodeWriter,
    "/datanodes/<string:datanode_id>/write",
    endpoint="datanode_writer",
    resource_class_kwargs={"logger": _logger},
)

api.add_resource(
    DataNodeList,
    "/datanodes",
    endpoint="datanodes",
    resource_class_kwargs={"logger": _logger},
)

api.add_resource(
    TaskResource,
    "/tasks/<string:task_id>",
    endpoint="task_by_id",
    resource_class_kwargs={"logger": _logger},
)
api.add_resource(TaskList, "/tasks", endpoint="tasks", resource_class_kwargs={"logger": _logger})
api.add_resource(
    TaskExecutor,
    "/tasks/submit/<string:task_id>",
    endpoint="task_submit",
    resource_class_kwargs={"logger": _logger},
)

api.add_resource(
    PipelineResource,
    "/pipelines/<string:pipeline_id>",
    endpoint="pipeline_by_id",
    resource_class_kwargs={"logger": _logger},
)
api.add_resource(
    PipelineList,
    "/pipelines",
    endpoint="pipelines",
    resource_class_kwargs={"logger": _logger},
)
api.add_resource(
    PipelineExecutor,
    "/pipelines/submit/<string:pipeline_id>",
    endpoint="pipeline_submit",
    resource_class_kwargs={"logger": _logger},
)

api.add_resource(
    ScenarioResource,
    "/scenarios/<string:scenario_id>",
    endpoint="scenario_by_id",
    resource_class_kwargs={"logger": _logger},
)
api.add_resource(
    ScenarioList,
    "/scenarios",
    endpoint="scenarios",
    resource_class_kwargs={"logger": _logger},
)
api.add_resource(
    ScenarioExecutor,
    "/scenarios/submit/<string:scenario_id>",
    endpoint="scenario_submit",
    resource_class_kwargs={"logger": _logger},
)

api.add_resource(
    CycleResource,
    "/cycles/<string:cycle_id>",
    endpoint="cycle_by_id",
    resource_class_kwargs={"logger": _logger},
)
api.add_resource(
    CycleList,
    "/cycles",
    endpoint="cycles",
    resource_class_kwargs={"logger": _logger},
)

api.add_resource(
    JobResource,
    "/jobs/<string:job_id>",
    endpoint="job_by_id",
    resource_class_kwargs={"logger": _logger},
)
api.add_resource(JobList, "/jobs", endpoint="jobs", resource_class_kwargs={"logger": _logger})
api.add_resource(
    JobExecutor,
    "/jobs/cancel/<string:job_id>",
    endpoint="job_cancel",
    resource_class_kwargs={"logger": _logger},
)


def load_enterprise_resources(api: Api):
    """
    Load enterprise resources.
    """

    if not _using_enterprise():
        return
    load_resources = _load_fct("taipy.enterprise.rest.api.views", "_load_resources")
    load_resources(api)


load_enterprise_resources(api)


@blueprint.before_app_first_request
def register_views():
    apispec.spec.components.schema("DataNodeSchema", schema=DataNodeSchema)
    apispec.spec.path(view=DataNodeResource, app=current_app)
    apispec.spec.path(view=DataNodeList, app=current_app)
    apispec.spec.path(view=DataNodeReader, app=current_app)
    apispec.spec.path(view=DataNodeWriter, app=current_app)

    apispec.spec.components.schema("TaskSchema", schema=TaskSchema)
    apispec.spec.path(view=TaskResource, app=current_app)
    apispec.spec.path(view=TaskList, app=current_app)
    apispec.spec.path(view=TaskExecutor, app=current_app)

    apispec.spec.components.schema("PipelineSchema", schema=PipelineSchema)
    apispec.spec.path(view=PipelineResource, app=current_app)
    apispec.spec.path(view=PipelineList, app=current_app)
    apispec.spec.path(view=PipelineExecutor, app=current_app)

    apispec.spec.components.schema("ScenarioSchema", schema=ScenarioSchema)
    apispec.spec.path(view=ScenarioResource, app=current_app)
    apispec.spec.path(view=ScenarioList, app=current_app)
    apispec.spec.path(view=ScenarioExecutor, app=current_app)

    apispec.spec.components.schema("CycleSchema", schema=CycleSchema)
    apispec.spec.path(view=CycleResource, app=current_app)
    apispec.spec.path(view=CycleList, app=current_app)

    apispec.spec.components.schema("JobSchema", schema=JobSchema)
    apispec.spec.path(view=JobResource, app=current_app)
    apispec.spec.path(view=JobList, app=current_app)
    apispec.spec.path(view=JobExecutor, app=current_app)

    apispec.spec.components.schema(
        "Any",
        {
            "description": "Any value",
            "nullable": True,
        },
    )

    if _using_enterprise():
        _register_views = _load_fct("taipy.enterprise.rest.api.views", "_register_views")
        _register_views(apispec)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from importlib.util import find_spec

if find_spec('taipy'):
    if find_spec("taipy.config"):
        from taipy.config import Config
        from taipy.config import Scope
        from taipy.config import Frequency

    if find_spec("taipy.core"):
        from taipy.core.taipy import (
            cancel_job,
            clean_all_entities,
            clean_all_entities_by_version,
            compare_scenarios,
            create_pipeline,
            create_scenario,
            delete,
            delete_job,
            delete_jobs,
            export_scenario,
            get,
            get_cycles,
            get_data_nodes,
            get_jobs,
            get_latest_job,
            get_parents,
            get_pipelines,
            get_primary,
            get_primary_scenarios,
            get_scenarios,
            get_tasks,
            set,
            set_primary,
            submit,
            subscribe_pipeline,
            subscribe_scenario,
            tag,
            unsubscribe_pipeline,
            unsubscribe_scenario,
            untag,
        )
        from taipy.core._core import Core
        from taipy.core.common.alias import CycleId, DataNodeId, JobId, PipelineId, ScenarioId, TaskId
        from taipy.core.cycle.cycle import Cycle
        from taipy.core.data.data_node import DataNode
        from taipy.core.job.job import Job
        from taipy.core.job.status import Status
        from taipy.core.pipeline.pipeline import Pipeline
        from taipy.core.scenario.scenario import Scenario
        from taipy.core.task.task import Task

    if find_spec('taipy.gui'):
        from taipy.gui import Gui

        if find_spec('taipy.enterprise') and find_spec('taipy.enterprise.gui'):
            from taipy.enterprise.gui import _init_gui_enterprise

            _init_gui_enterprise(Gui)

    if find_spec('taipy.rest'):
        from taipy.rest import Rest

    if find_spec('taipy._run'):
        from taipy._run import _run as run
